# Copyright 2024 Mohamed Gaber

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path
from typing import Iterable

import click

from .nl2bench import verilog_netlist_to_bench


@click.command()
@click.option(
    "-o", "--output", default=None, help="If not set, replacing .v with .bench"
)
@click.option(
    "-l",
    "--lib-file",
    "lib_files",
    multiple=True,
    type=click.Path(file_okay=True, exists=True),
)
@click.option(
    "-b",
    "--bypassing",
    "bypassed_ports",
    multiple=True,
    type=str,
)
@click.option(
    "--msb-first/--lsb-first",
    default=False,
    type=bool,
)
@click.argument(
    "netlist_in",
    type=click.Path(file_okay=True, exists=True),
)
def cli(
    output: str,
    netlist_in: str,
    bypassed_ports: Iterable[str],
    lib_files: Iterable[str],
    msb_first: bool,
):
    if output is None:
        output = netlist_in
        if output.endswith(".v"):
            output = output[:-2]
        output = f"{output}.bench"
    with open(output, "w", encoding="utf8") as f:
        verilog_netlist_to_bench(
            Path(netlist_in),
            lib_files,
            f,
            bypass_ios=set(bypassed_ports),
            msb_first=msb_first,
        )
    print(f"Successfully saved to {output}.")


if __name__ == "__main__":
    cli()



# Copyright 2024 Mohamed Gaber

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from io import TextIOWrapper
from pathlib import Path
from typing import Iterable, Dict, Optional, Set

from . import nl_parser
from .cell import Cell


def n(identifier):
    if identifier.startswith("\\") and identifier.endswith(" "):
        identifier = identifier[1:-1]
    return identifier


def transform_function_rec(
    current_output: str,
    current_function,
    inst: nl_parser.Instance,
    f: TextIOWrapper,
    depth=0,
):
    argi = 0
    current_output = n(current_output)

    def handle_argument(arg):
        nonlocal argi
        if isinstance(arg, tuple):
            intermediate = current_output + f".{argi}"
            argi += 1
            transform_function_rec(intermediate, arg, inst, f, depth + 1)
            return intermediate
        else:
            assert (
                arg in inst.io
            ), f"Unrecognized input port {arg} in {inst.name} of kind {inst.kind}"
            return n(inst.io[arg])

    if isinstance(current_function, str):
        if current_function == "0":
            print(
                f"{current_output} = VSS()",
                file=f,
            )
        elif current_function == "1":
            print(
                f"{current_output} = VDD()",
                file=f,
            )
        else:
            print(
                f"{current_output} = BUF({handle_argument(current_function)})",
                file=f,
            )
    elif current_function[0] == "!":
        print(f"{current_output} = NOT({handle_argument(current_function[1])})", file=f)
    elif current_function[0] == "&":
        print(
            f"{current_output} = AND({handle_argument(current_function[1])} , {handle_argument(current_function[2])})",
            file=f,
        )
    elif current_function[0] == "^":
        print(
            f"{current_output} = XOR({handle_argument(current_function[1])} , {handle_argument(current_function[2])})",
            file=f,
        )
    elif current_function[0] == "+":
        print(
            f"{current_output} = OR({handle_argument(current_function[1])} , {handle_argument(current_function[2])})",
            file=f,
        )
    else:
        raise ValueError(
            f"Unknown lib function {current_function} of type {type(current_function)} in {inst}."
        )


def io_to_bench(port: nl_parser.Port, f: TextIOWrapper, msb_first: bool = False):
    if port.msb is None:
        print(f"{port.direction.upper()}({n(port.name)})", file=f)
    else:
        if msb_first:
            frm, to = max(port.msb, port.lsb), min(port.lsb, port.msb)
            i = frm
            while i >= to:
                print(f"{port.direction.upper()}({n(port.name)}[{i}])", file=f)
                i -= 1
        else:
            frm, to = min(port.msb, port.lsb), max(port.lsb, port.msb)
            i = frm
            while i <= to:
                print(f"{port.direction.upper()}({n(port.name)}[{i}])", file=f)
                i += 1


def statements_to_bench(inst: nl_parser.Instance, base: Cell, f: TextIOWrapper):
    for output, function in base.outputs.items():
        if hooked := inst.io.get(output):
            # HACK FOR SKY130
            if function is None and inst.kind.startswith("sky130_fd_sc_hd__dlclkp_"):
                function = "1"
            transform_function_rec(hooked, function, inst, f)


def netlist_to_bench(
    netlist: nl_parser.Netlist,
    cells: Dict[str, Cell],
    f: TextIOWrapper,
    bypass_ios: Optional[Set[str]] = None,
    msb_first: bool = False,
):
    bypass_ios = bypass_ios or set()
    print(f"# module {netlist.module}", file=f)
    print("# Automatically generated by nl2bench. Do not modify.", file=f)
    for port in netlist.ports.values():
        if port.name in bypass_ios:
            continue
        io_to_bench(port, f, msb_first)
    for inst in netlist.instances:
        assert (
            inst.kind in cells
        ), f"Unknown cell {inst.kind}- are you sure the design is combinational?"
        base = cells[inst.kind]
        statements_to_bench(inst, base, f)
    for asst in netlist.assignments:
        if asst[1] == 0:
            print(f"{n(asst[0])} = VSS()", file=f)
        elif asst[1] == 1:
            print(f"{n(asst[0])} = VDD()", file=f)
        else:
            print(f"{n(asst[0])} = BUF({n(asst[1])})", file=f)


def verilog_netlist_to_bench(
    netlist_in: Path,
    lib_files: Iterable[str],
    f: TextIOWrapper,
    bypass_ios: Optional[Set[str]] = None,
    msb_first: bool = False,
):
    cells: Dict[str, Cell] = functools.reduce(
        lambda acc, path: acc.update(Cell.from_lib_file(path)) or acc,
        lib_files,
        {},
    )

    netlist = nl_parser.parse(netlist_in)

    netlist_to_bench(netlist, cells, f, bypass_ios, msb_first)



# Copyright 2024 Mohamed Gaber

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

from libparse import LibertyParser, LibertyAst
import frozendict

from .lib_fn_parser import parse


def lib_group_as_dict(entry) -> dict:
    retval = {
        "args": tuple(entry.args),
    }
    for child in entry.children:
        value = child.value
        if isinstance(value, LibertyAst):
            value = lib_group_as_dict(value)
        retval[child.id] = child.value
    return retval


@dataclass(frozen=True)
class TestInfo:
    testing_ff: frozendict.frozendict
    sco: str
    sci: str
    sce: str


@dataclass
class Cell:
    inputs: List[str]
    outputs: Dict[str, Optional[tuple]]
    inouts: List[str]
    register_info: Optional[Tuple[str, frozendict.frozendict]]
    test_info: Optional[TestInfo]
    raw: frozendict.frozendict

    @staticmethod
    def _from_ast(cell_ast) -> "Cell":
        name = cell_ast.args[0]
        raw = lib_group_as_dict(cell_ast)

        inputs = []
        outputs = {}
        inouts = []
        register_info = None
        test_info = None
        for cell_element in cell_ast.children:
            if cell_element.id in ["ff", "latch"]:
                register_info = (
                    cell_element.id,
                    frozendict.deepfreeze(lib_group_as_dict(cell_element)),
                )
        for cell_element in cell_ast.children:
            if cell_element.id == "test_cell":
                test_info_dict = {}
                for test_element in cell_element.children:
                    if test_element.id == "ff":
                        test_info_dict["testing_ff"] = frozendict.deepfreeze(
                            lib_group_as_dict(test_element)
                        )
                    elif test_element.id == "pin":
                        pin_info = lib_group_as_dict(test_element)
                        pin_name = test_element.args[0]
                        if sigtype := pin_info.get("signal_type"):
                            if sigtype == "test_scan_out":
                                test_info_dict["sco"] = pin_name
                            elif sigtype == "test_scan_in":
                                test_info_dict["sci"] = pin_name
                            elif sigtype == "test_scan_enable":
                                test_info_dict["sce"] = pin_name
                test_info = TestInfo(**test_info_dict)
            if cell_element.id == "pin":
                function = None
                direction = None
                for pin_element in cell_element.children:
                    if pin_element.id == "direction":
                        direction = pin_element.value
                    if pin_element.id == "function":
                        function = pin_element.value

                if direction == "input":
                    inputs.append(cell_element.args[0])
                elif direction == "output":
                    if fn := function:
                        outputs[cell_element.args[0]] = parse(fn)
                    else:
                        outputs[cell_element.args[0]] = None
                else:
                    inouts.append(cell_element.args)

        return name, Cell(inputs, outputs, inouts, register_info, test_info, raw)

    @staticmethod
    def from_lib_file(path: str) -> Dict[str, "Cell"]:
        cells: Dict[str, Cell] = {}

        parser = LibertyParser(open(path))
        ast = parser.ast
        cells: Dict[str, Cell] = {}
        for entry in ast.children:
            if entry.id == "cell":
                name, cell = Cell._from_ast(entry)
                cells[name] = cell
        return cells



# Copyright 2024 Mohamed Gaber

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from antlr4 import InputStream, CommonTokenStream

from _nl2bench_antlr4_liblogic.logicListener import logicListener
from _nl2bench_antlr4_liblogic.logicParser import logicParser
from _nl2bench_antlr4_liblogic.logicVisitor import logicVisitor
from _nl2bench_antlr4_liblogic.logicLexer import logicLexer


class _Listener(logicListener):
    def syntaxError(*args, **kwargs):
        raise Exception(f"{args} {kwargs}")


class _Visitor(logicVisitor):
    def visitFunction(self, ctx: logicParser.FunctionContext):
        children = [str(c) for c in ctx.children]
        if len(ctx.children) == 1:  # name
            return children[0]
        if children[0] == "(":
            return self.visit(ctx.children[1])
        if children[0] == "!":
            return ("!", self.visit(ctx.children[1]))
        if len(children) == 2:
            # guaranteed to be and at this point
            return ("&", self.visit(ctx.children[0]), self.visit(ctx.children[1]))
        operator = children[1]
        if operator == "|":
            operator = "+"
        return (operator, self.visit(ctx.children[0]), self.visit(ctx.children[2]))


def parse(function: str):
    listener = _Listener()
    stream = InputStream(function)

    lexer = logicLexer(stream)
    lexer.addErrorListener(listener)
    token_stream = CommonTokenStream(lexer)
    token_stream.fill()
    parser = logicParser(token_stream)
    parser.addErrorListener(listener)

    tree = parser.function()

    return _Visitor().visit(tree)



# Copyright 2024 Mohamed Gaber, Efabless Corporation

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Code partially adapted from Verilog Backend of:
#
# yosys -- Yosys Open SYnthesis Suite
#
# Copyright (C) 2012  Claire Xenia Wolf <claire@yosyshq.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from pathlib import Path
import shlex
from typing import List, Dict, Optional, Tuple
from dataclasses import dataclass, field


from pyosys import libyosys as ys


@dataclass
class Instance:
    name: str
    kind: str
    io: Dict[str, str] = field(default_factory=lambda: {})


@dataclass
class Port:
    name: str
    direction: str = "Unknown"
    msb: Optional[int] = None
    lsb: Optional[int] = None


@dataclass
class Netlist:
    module: str = ""
    ports: Dict[str, Port] = field(default_factory=lambda: {})
    instances: List[Instance] = field(default_factory=lambda: [])
    assignments: List[Tuple[str, str]] = field(default_factory=lambda: [])


def _clean_str(id: ys.IdString):
    string_id = id.str()
    if string_id.startswith("\\"):
        string_id = string_id[1:]
    return string_id


def _dump_sigchunk(chunk: ys.SigChunk):
    assert chunk.wire, "constants not supported"
    assert chunk.width == 1, "splitnets did not split this mf net"
    wire_name = _clean_str(chunk.wire.name)
    if chunk.width == chunk.wire.width:
        return wire_name
    else:
        if chunk.wire.upto:
            return f"{wire_name}[{chunk.wire.width - (chunk.offset - 1) + chunk.wire.start_offset}]"
        else:
            return f"{wire_name}[{chunk.offset + chunk.wire.start_offset}]"


def _dump_sigbit(bit: ys.SigBit):
    if bit.is_wire():
        if bit.wire.width == 1:
            return _clean_str(bit.wire.name)
        else:
            return f"{_clean_str(bit.wire.name)}[{bit.offset}]"
    else:
        if bit.data == ys.State.S1:
            return 1
        elif bit.data == ys.State.S0:
            return 0
        else:
            assert "unknown constants not supported"


def parse(verilog_netlist: Path):
    d = ys.Design()
    ys.run_pass(f"read_verilog {shlex.quote(str(verilog_netlist))}", d)
    ys.run_pass("hierarchy -auto-top", d)
    ys.run_pass("flatten", d)
    ys.run_pass("splitnets", d)
    ys.run_pass("opt_clean -purge", d)
    ys.run_pass("setundef -zero -undriven", d)

    ys_module = d.top_module()

    ports: Dict[str, Port] = {
        # just to keep declaration order
        _clean_str(port): Port(name=_clean_str(port))
        for port in ys_module.ports
    }

    for wire_idstr in ys_module.wires_:
        wire = ys_module.wire(wire_idstr)
        if wire.port_id == 0:
            continue
        port_name = _clean_str(wire.name)

        msb = None
        lsb = None
        upto = wire.upto == 1
        offset = wire.start_offset
        width = wire.width
        if width > 1:
            msb = offset + width - 1
            lsb = offset
            if upto:
                msb, lsb = lsb, msb
        ports[port_name] = Port(
            port_name,
            "input" if wire.port_input else "output",
            msb,
            lsb,
        )

    assignments: List[Tuple[str, str]] = []
    for conn_from, conn_to in ys_module.connections():
        for bit_from, bit_to in zip(conn_from.bits(), conn_to.bits()):
            assignments.append((_dump_sigbit(bit_from), _dump_sigbit(bit_to)))

    instances: List[Instance] = []
    for cell_idstr in ys_module.cells_:
        cell = ys_module.cell(cell_idstr)
        instance_io = {}
        for port_idstr, spec in cell.connections().items():
            instance_io[_clean_str(port_idstr)] = _dump_sigchunk(spec.as_chunk())
        cell_name = _clean_str(cell.name)
        cell_type = _clean_str(cell.type)
        instances.append(Instance(cell_name, cell_type, instance_io))

    return Netlist(_clean_str(ys_module.name), ports, instances, assignments)
